//
//  PlaygroundViewController.swift
//  SkoogAPI
//
//  Created by Keith Nagle on 02/08/2016.
//  Copyright © 2016 Skoogmusic Ltd. All rights reserved.
//

import UIKit
import SpriteKit
import CoreBluetooth
import CoreAudioKit

public typealias Index = Int

open class PlaygroundViewController: UIViewController, bleMessagesDelegate, SkoogSceneDelegate, SkoogDelegate {
    
//    public var showCircle = true
    public var changeAlpha = false
    public var growCircle = false
    public var showRipples = false
    public var playNotes = true
    public var changeColor = true
    public var showPressLabels = false
    public var showSqueezeLabel = false

    let playground      = self
	
	/// Instrument
    public var instrument  : Instrument = Instrument()
    public var instType	: SynthType = .sineWave

    
    
    public var notes : [UInt] = [58,65,68,70,73]
    

    public var circle = SkoogScene(size: CGRect(x: 0,
                                             y:	0,
                                             width: 512,
                                             height: 768).size
    )
    
    public let skoogSKView = SKView(frame: CGRect(x: 0,
                                           y:	0,
                                           width: 512,
                                           height: 768)
    )
    
    public var uiButton    = UIButton(type: .system) as UIButton
    
    let cabtViewController	=	CABTMIDICentralViewController.init()
    let cabtContainer		=	UIView()
    
    var bluetooth : BLEMessages?
    public var skoogConnected : Bool = false
    var newConnection : Bool = false
    var connectedSkoog : CBPeripheral? = nil
    var readyToPlay : Bool = false

    
    public let skoog    = Skoog.sharedInstance
    
    public let red = Side()
    public let blue = Side()
    public let yellow = Side()
    public let green = Side()
    public let orange = Side()
    public var currentSide = Side()
    
    public init() {
        super.init(nibName: nil, bundle: nil)
        // Do any additional setup after loading the view, typically from a nib.
    }
    
    public required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
    }
    
    open override func viewDidLoad() {
		super.viewDidLoad()
        skoog.setPolyMode(active: false)
        
        red.index = 0
        blue.index = 1
        yellow.index = 2
        green.index = 3
        orange.index = 4
        
        skoog.delegate = self
        
        bluetooth = BLEMessages()
        bluetooth?.delegate = self
        
        instrumentSetup() // change instrument soundfont
        
        // Our feedback circle display
        self.view.addSubview(skoogSKView)
        skoogSKView.ignoresSiblingOrder = true
        circle.scaleMode = .resizeFill
        circle.sceneDelegate = self
        skoogSKView.presentScene(circle)
        
        // adjust sensitivity for each side of the Skoog. Value range is 0 to 12
        for i in 0 ... 4 {
            setResponse(index: i, value: 6.0)
        }
        
        // set up the CABTMIDICentralViewController and a UIView to display it in
        cabtViewController.view.frame	=	CGRect(x:		0,
                                     	 	       y:		0,
                                     	 	       width:	0.5 * self.view.frame.size.width,
                                     	 	       height:	0.25 * self.view.frame.size.height)
        
        self.addChildViewController(cabtViewController)
        self.view.addSubview(cabtContainer)
        cabtViewController.tableView.backgroundColor	= .clear
        
        cabtContainer.addSubview(cabtViewController.view)
        cabtContainer.frame				=	CGRect(x:		0,
                           				 	       y:		0,
                           				 	       width:	self.view.frame.size.width,
                           				 	       height:	self.view.frame.size.height)
        
        // A button to represent a connected Skoog. tapping the button presents the 
        // CABTMIDICentralViewController if its been hidden.
        uiButton.frame  = CGRect(x: 468, y: 0, width: 44, height: 44)
        uiButton.setImage(#imageLiteral(resourceName: "skios_skoog_baricon.png"), for: .normal)
        uiButton.tintColor = .darkGray
        self.view.addSubview(uiButton)
        uiButton.addTarget(self, action: #selector(showCABTController), for: .touchUpInside)
        
        // A bit of a fudge. The timer corrects a display issue where the CABTMIDICentralViewController
        // might display part of itself "offscreen"
        Timer.scheduledTimer(timeInterval:0.1, target: self, selector: #selector(skoogSearchTask), userInfo: nil, repeats: false)
    }
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Skoog Functions
/////////////////////////////////////////////////////////////////////////
    
    
    public func skoogSearchTask() {
        skoog.searchForSkoog()
        // if we have a BLE connection to Skoog this will trigger a calibration
        // If not, nothing will happen, it will fail gracefully.
        touchUp() //perform the calibration routine
    }
    
    public func setResponse(index: Int, value: Double) {
        skoog.sides[index].response = value
    }
    
    public func getReadyToPlay() -> Bool {
        return readyToPlay
    }
    
    public func setReadyToPlay(value: Bool? = nil){
        if let val = value {
            if val {
                if skoog.skoogConnected {
                    uiButton.tintColor = circle.colors[3]
                }
            }
            else{
                if skoog.skoogConnected {
                    uiButton.tintColor = circle.colors[4]
                }
            }
            self.readyToPlay = val
        }
    }
    
    public func setReadyToPlay(){
        if skoog.skoogConnected {
            self.readyToPlay = true
            uiButton.tintColor = circle.colors[3]
        }
    }
    
    public func calibrateTask() {
        bluetooth?.calibrate()
        Timer.scheduledTimer(timeInterval:0.8, target: self, selector: #selector(setReadyToPlay), userInfo: nil, repeats: false)
        //        self.readyToPlay = true
    }
    
    public func skoogConnectionStatus(connected: Bool) {
        if connected == true && skoog.skoogConnected == true{
            self.hideCABTController()
            touchUp()
        }
        else {
            setReadyToPlay(value:false) //this has to happen beofew showing the CABT
            self.showCABTController()
        }
    }
    
    open override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    
/////////////////////////////////////////////////////////////////////////
// MARK: - Bluetooth Functions
/////////////////////////////////////////////////////////////////////////
    
    public func showCABTController(){
        cabtContainer.frame				=	CGRect(x:		0,
                           				 	       y:		0,
                           				 	       width:	self.view.frame.size.width,
                           				 	       height:	self.view.frame.size.height)
        cabtViewController.view.frame	=	CGRect(x:		0,
                                     	 	       y:		0,
                                     	 	       width:	cabtContainer.frame.size.width,
                                     	 	       height:	cabtContainer.frame.size.height)
        uiButton.tintColor = .darkGray
    }
    
    public func hideCABTController() {
        cabtContainer.frame				=	CGRect(x:		0,
                           				 	       y:		0,
                           				 	       width:	0,
                           				 	       height:	0)
        cabtViewController.view.frame	=	CGRect(x:		0,
                                     	 	       y:		0,
                                     	 	       width:	0,
                                     	 	       height:	0)
        uiButton.tintColor = circle.colors[4]
    }

    
    
    
    
    
//    public func circleShow() {
//        self.showCircle = true
//        circle.drawCircle = true
//    }
    
    
    
    
    public func hideCircle() {
        //get rid of current ripple
        if let theripple = circle.ripple {
            theripple.run(SKAction.removeFromParent())
        }
        // get rid of circle
        if let thecircle = circle.circle {
            thecircle.run(SKAction.removeFromParent())
            circle.circle = nil
        }
        //Make new smaller ripple
        circle.ripple = circle.makeCircle(location:CGPoint(x:256,y:384), radius: 10, color: .clear)
       
        
//        self.showCircle = false
    }
    
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Alpha Functions
/////////////////////////////////////////////////////////////////////////
    
    public func alphaOn() {
        self.changeAlpha = true
    }
    
    public func alphaOff() {
        self.changeAlpha = false
    }

    public func alpha(value: Double){
        circle.changeAlpha(strength: value)
    }
    
    public func alpha(side: Side, value: Double){
        circle.changeAlpha(side: side.index, strength: value)
    }

    
/////////////////////////////////////////////////////////////////////////
// MARK: - Ripple Functions
/////////////////////////////////////////////////////////////////////////
    
    public func rippleOn() {
        showRipples = true
    }
    
    public func ripple(setTime: Double) {
        if setTime <= 0.0 {
            circle.durationMultiplier = 0.001
        }
        else {
            circle.durationMultiplier = setTime
        }
    }
	
/**
	Triggers an expanding circle in the liveView.
	
	- parameters:
	- color: lets the ripple know what color to be (optional).  Defaults to `.white` if unspecified.
	- size: tells it how big to be when it is created (optional). Defaults to `1.0` if unspecified.
	- time: tells the ripple how quickly to expand. Defaults to `1.0` if unspecified.
*/
    public func ripple(color: UIColor? = nil, size: Double? = nil, time: Double? = nil) {
        if let t = time {
            if t <= 0.0 {
                circle.durationMultiplier = 0.001
            }
            else {
                circle.durationMultiplier = t
            }
        }
        var c = color
		if c == nil {
            c = SKColor(colorLiteralRed: 1.0, green: 1.0, blue: 1.0, alpha: 1.0)   // white
		}
        
        var s = size
		if s == nil {
            s = 0.025
        }
        circle.startRipple(color: c!, size: s!)
    }
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Grow Functions
/////////////////////////////////////////////////////////////////////////
    
	public func grow(side: Side, value: Double){
		circle.changeSize(side: side.index, size: value)
	}
    
    public func grow(value: Double){
        circle.changeSize(size: value)
    }
	
    public func grow() {
        self.growCircle = true
    }
    
    public func dontGrow() {
        self.growCircle = false
    }
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Glow Functions
/////////////////////////////////////////////////////////////////////////
    
    public func glow(value: Double){
        circle.setGlow(amount: value)
    }
    
    
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Touchscreen Functions
/////////////////////////////////////////////////////////////////////////
    
    
    public func touchUp() {
        setReadyToPlay(value:false)
//        self.readyToPlay = false
        bluetooth?.findSkoog(shouldConnect: true)
        Timer.scheduledTimer(timeInterval:0.05, target: self, selector: #selector(calibrateTask), userInfo: nil, repeats: false)
    }
    
    public func touchDown() {
    }
    
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Instrument setup methods
/////////////////////////////////////////////////////////////////////////
    
    // set a default instrument sound.
    public func instrumentSetup() {
        setInstrument(.sineWave)
    }
	
	/**
	Set the instrument for the Skoog.
	- parameters:
		- name: instrument name.
     
          Available instruments:
          .acid
          .sineWave
          .superSaw
          .candyBee
          .solarWind
          .fmModulator
	*/
    public func setInstrument(_ name: SynthType) {
        instrument.type = name
        instrument.loadMIDISynthSoundFont(type: name)
    }
/////////////////////////////////////////////////////////////////////////
// MARK: - Note setup methods
/////////////////////////////////////////////////////////////////////////

	/**
	Helper function to set the musical note for each side of the Skoog. Notes can be set individually or all together. Here is how to set each side to play the first 5 notes of the C-major scale:
     
        setNotes(red:    60,
                 blue:   62,
                 yellow: 64,
                 green:  65,
                 orange: 67)
     
	 - parameters:
		- red: Set the MIDI number for the red side (optional).
		- blue: Set the MIDI number for the blue side (optional).
		- yellow: Set the MIDI number for the yellow side (optional).
		- green: Set the MIDI number for the green side (optional).
		- orange: Set the MIDI number for the orange side (optional).
	 */
    public func setNotes(red: UInt? = nil, blue: UInt? = nil, yellow: UInt? = nil, green: UInt? = nil, orange: UInt? = nil){
        if let Red = red {
            notes[0] = Red
        }
        
        if let Blue = blue {
            notes[1] = Blue
        }
        
        if let Yellow = yellow {
            notes[2] = Yellow
        }
        
        if let Green = green {
            notes[3] = Green
        }
        
        if let Orange = orange {
            notes[4] = Orange
        }
    }
    
/////////////////////////////////////////////////////////////////////////
// MARK: - MIDI methods
/////////////////////////////////////////////////////////////////////////

    public func noteOn(sideIndex: Index, velocity:Double) {
        if readyToPlay == true{
            instrument.noteOn(noteNum: notes[sideIndex],
                              velocity: UInt32(48.0 + velocity*79.0),
                              channel: UInt32(sideIndex))
        }
    }
    
    public func noteOff(sideIndex: Int) {
        instrument.noteOff(noteNum: notes[sideIndex],
                           channel: UInt32(sideIndex))
    }
    
    public func volume(sideIndex: Int, value:Double){
        instrument.volume(value: UInt32(Float(value) * 127.0),
                          channel: UInt32(sideIndex)) //offset by 2: MPE note channels start at 2
    }
    
    
    public func modulate(sideIndex: Int, value:Double){
        instrument.modulate(value: UInt32(Float(value) * 127.0),
                            channel: UInt32(sideIndex)) //offset by 2: MPE note channels start at 2
    }
    
    public func pan(sideIndex: Int, value:Double){
        instrument.pan(value: UInt32(Float(value) * 127.0),
                       channel: UInt32(sideIndex)) //offset by 2: MPE note channels start at 2
    }
    
    public func expression(sideIndex: Int, value:Double){
        instrument.expression(value: UInt32(Float(value) * 127.0),
                              channel: UInt32(sideIndex)) //offset by 2: MPE note channels start at 2
    }
    
    public func pitchBend(sideIndex: Int, value:Double){
        instrument.pitchBend(bend: UInt32(Float(1 + value) * 8192.0),
                             channel: UInt32(sideIndex))
    }
    
    public func pitchBendUp(sideIndex: Index, value:Double){
        
        instrument.pitchBend(bend: UInt32(Float(1 + value) * 8192.0),
                             channel: UInt32(sideIndex))
    }
    
    public func pitchBendDown(sideIndex: Int, value:Double){
        instrument.pitchBend(bend: UInt32(Float(1 - value) * 8192.0),
                             channel: UInt32(sideIndex))
    }
    
    public func pitchBendRange(range:Int){
        for i in 0 ... 4 {
            instrument.pitchRange(range: UInt32(range),
                                  channel: UInt32(i)) //offset by 2: MPE note channels start at 2
        }
    }
    
    public func channelPressure(sideIndex: Int, value:Double){
        instrument.channelPressure(pressure: UInt32(Float(value) * 127.0),
                                   channel: UInt32(sideIndex))
    }

	
	public func afterTouch(sideIndex: Int, value:Double){
		instrument.afterTouch(note:		notes[sideIndex],
		                      touch:	UInt32(Float(value) * 127.0),
		                      channel:	UInt32(sideIndex))
	}
    
/////////////////////////////////////////////////////////////////////////
// MARK: - Delegate methods - overridden in Contents.swift
/////////////////////////////////////////////////////////////////////////
    
    open func peak(side: Side, peak: Double) {
        // handle global setup
        if self.playNotes {
            noteOn(sideIndex:side.index, velocity: peak)
        }
        if self.changeColor {
            circle.changeCircleColor(side: side.index, size: peak)
        }
        if self.showPressLabels {
            circle.labelFade(side:side, value: peak)
        }
        if self.showRipples {
            ripple(color: side.color, size: peak)
        }
        
    }
    
    open func continuous(side: Side) {
        // handle global setup
        if self.changeAlpha {
            circle.changeAlpha(side: side.index, strength: side.rawValue)
        }
        if self.growCircle {
            circle.changeSize(side: side.index, size: side.rawValue)
        }
        if self.showSqueezeLabel {
            circle.label?.text = "Squeeze = " + String(format:"%0.2f", side.rawValue)
            circle.label?.alpha = CGFloat(side.rawValue)
        }
    }
    
    open func release(side: Side) {
        // handle global setup
        if self.playNotes {
            noteOff(sideIndex:side.index)
        }
        if self.changeColor {
            circle.changeCircleColor(side: side.index, size: 0.0)
        }
    }
}
