//
//  Skoog.swift
//  Skoog
//
//  Created by Keith Nagle on 10/06/2016.
//  Copyright © 2016 Skoogmusic Ltd. All rights reserved.
//
import Foundation
import Darwin
import CoreMIDI
import UIKit

public protocol SkoogDelegate: class {
    func peak(side: Side, peak: Double)
    func continuous(side: Side)
    func release(side: Side)
    func skoogConnectionStatus(connected: Bool)
}


public class Side: NSObject {
    public var active : Bool = false
    public var color : UIColor = UIColor(red: 0.0, green: 0.0, blue: 0.0, alpha: 0.0)
    public var rawValue : Double = 0
    public var value : Double = 0
    public var peak : Peak? = Peak.init()
    public var index : Int = 0 // perhaps not needed but useful for identifying by index for now
    public var response : Double = 6.0
    public var angle : Double = 0.0
    public var name : String = ""
    public var blend_in_xy : Double = 0.0
    public var blend_out_xy : Double = 0.0
    public override init() {
        super.init()
    }
}

// The "singleton" instance
public let SkoogInstance = Skoog() 

public class Skoog: NSObject {
    
    public weak var delegate:SkoogDelegate?
    
    enum Zone: Int {
        case red_side = 0,      // 1
        blue_side,              // 2
        yellow_side,            // 3
        green_side,             // 4
        orange_side,            // 5
        red_blue,               // 6
        blue_yellow,            // 7
        yellow_green,           // 8
        red_green,              // 9
        orange_red,             // 10
        orange_red_blue,        // 11
        orange_blue,            // 12
        orange_blue_yellow,     // 13
        orange_yellow,          // 14
        orange_yellow_green,    // 15
        orange_green,           // 16
        orange_red_green,       // 17
        none
    }

    var activeZones : [[Zone]] = [[.red_side, .none, .none],
                                 [.blue_side, .none, .none],
                                 [.yellow_side, .none, .none],
                                 [.green_side, .none, .none],
                                 [.orange_side, .none, .none],
                                 [.red_side, .blue_side, .none],
                                 [.blue_side, .yellow_side, .none],
                                 [.yellow_side, .green_side, .none],
                                 [.red_side, .green_side, .none],
                                 [.orange_side, .red_side, .none],
                                 [.orange_side, .red_side, .blue_side],
                                 [.orange_side, .blue_side, .none],
                                 [.orange_side, .blue_side, .yellow_side],
                                 [.orange_side, .yellow_side, .none],
                                 [.orange_side, .yellow_side, .green_side],
                                 [.orange_side, .green_side, .none],
                                 [.orange_side, .red_side, .green_side],
                                 [.none, .none, .none]];
    
    public class var sharedInstance:Skoog {
        return SkoogInstance
    }
    
    var status = OSStatus(noErr)
    var midiClient = MIDIClientRef()
    var outputPort = MIDIPortRef()
    var inputPort = MIDIPortRef()
    var virtualSourceEndpointRef = MIDIEndpointRef()
    var virtualDestinationEndpointRef = MIDIEndpointRef()
    var midiInputPortref = MIDIPortRef()
    
    public var red : Side
    public var blue : Side
    public var yellow : Side
    public var green : Side
    public var orange : Side
    var sides : [Side]
    
    public var polyMode = false
    
    var R1 : Double = 0.0
    var R2 : Double = 0.0
    var T1 : Double = 0.0
    var T2 : Double = 0.0
    var Z3 : Double = 0.0
    var ZC : Double = 0.0
    var DX : Double = 0.0
    var DY : Double = 0.0
    var Rxy : Double = 0.0
    var Rxyz : Double = 0.0
    public var Txy : Double = 0.0
    var dT : Double = 0.0
    var Tz : Double = 0.0
    var R_zoom : Double = 0.0
    var Z_zoom : Double = 0.0
    
    var Rxyz_old : Double = 0.0;
    var Rxy_old : Double = 0.0;
    var R1_old : Double = 0.0;
    var R2_old : Double = 0.0;
    var T1_old : Double = 0.0;
    var T2_old : Double = 0.0;
    var Tz_old : Double = 0.0;
    var Z3_old : Double = 0.0;
    
    var lastZone : Zone = .none
    var zone : Zone = .none
    var param2 : Double = 0
    var param3 : Double = 0
    var param4 : Double = 0
    
    var threshold : Double = 0
    var z_threshold : Double = 0
    var thresholdZoom : Double = 0
    var z_thresholdZoom : Double = 0
    var XYcone : Double = 15.0
    var ZconeZ : Double = 11.5
    var ZconeXY : Double = 22.5
    var coneAngle : Double = 0
    var redBlueAngle : Double = 0
    var blueYellowAngle : Double = 0
    var yellowGreenAngle : Double = 0
    var redGreenAngle : Double = 0
    var bend_note : Double = 0
    var bend_out : Double = 0
    var bend_in : Double = 0
    var blend_out_xy : Double = 0
    var blend_in_xy : Double = 0
    var blend_out_z : Double = 0
    var blend_in_z : Double = 0
    
    var Tz_corrected : Double = 0
    var Z3_corrected : Double = 0
    var blocking = false
    public var skoogConnected = false
    var currenttime = NSDate()
    var lasttime = NSDate()
    
    override init() {
        self.red = Side()
        self.blue = Side()
        self.yellow = Side()
        self.green = Side()
        self.orange = Side()
        
        self.red.index = 0
        self.blue.index = 1
        self.yellow.index = 2
        self.green.index = 3
        self.orange.index = 4
        
        self.red.name = "red"
        self.blue.name = "blue"
        self.yellow.name = "yellow"
        self.green.name = "green"
        self.orange.name = "orange"
        
        self.sides = [red, blue, yellow, green, orange]
        
        self.red.color      =   UIColor(colorLiteralRed: 218.0/255.0, green: 60.0/255.0, blue: 0.0, alpha: 1.0)
        self.blue.color     =   UIColor(colorLiteralRed: 55.0/255.0, green: 127.0/255.0, blue: 178.0/255.0, alpha: 1.0)
        self.yellow.color   =   UIColor(colorLiteralRed: 250.0/255.0, green: 217.0/255.0, blue: 0.0, alpha: 1.0)
        self.green.color    =   UIColor(colorLiteralRed: 163.0/255.0, green: 203.0/255.0, blue: 0.0, alpha: 1.0)
        self.orange.color   =   UIColor(colorLiteralRed: 249.0/255.0, green: 154.0/255.0, blue: 0.0, alpha: 1.0)
        
        
        super.init()
        
        self.R1 = 0
        self.T1 = 180
        self.R2 = 0
        self.T2 = 180
        self.Z3 = 0
        self.DX = 64
        self.DY = 64
        self.Rxy = 0
        self.Rxyz = 0
        self.Txy = 180
        self.dT = 0
        self.Tz = 90
        self.threshold = 0.0014 // was 0.02
        self.z_threshold = 0.0014
        self.thresholdZoom = 0.0014
        self.z_thresholdZoom = 0.0014 // 0.018
        self.R_zoom = 1.0
        self.Z_zoom = 1.0
        
        self.XYcone = 16 //these values "fill" the space - temporary until cone of shame implememted
        self.ZconeZ = 16.5
        self.ZconeXY = 10.0
        self.zone = .none
        self.lastZone = .none
        self.param3 = 0.0
        self.param4 = 0.0
        
        setPolyMode(active:self.polyMode)
//        changeXYcone(XYcone)
        searchForSkoog()
        
        status = MIDIClientCreateWithBlock("com.skoogmusic.myMIDIClient" as CFString, &midiClient, MyMIDINotifyBlock)
        if status != noErr {
            print("Error creating midi client : \(status)")
        }
        else{
            print("Created midi client : \(status)")
        }

        MIDISourceCreate(midiClient, "MidiTest_VirtualPort" as CFString, &virtualSourceEndpointRef)
    }
// MARK: MIDI methods
    func MyMIDINotifyBlock(midiNotification: UnsafePointer<MIDINotification>) {
        let notification = midiNotification.pointee
        switch (notification.messageID) {
        case .msgSetupChanged:
            print("Setup changed!")
            searchForSkoog()
        case .msgObjectAdded:
            print("Object added!")
            if !skoogConnected {
                searchForSkoog()
            }
        case .msgObjectRemoved:
            print("Object removed!")
            if skoogConnected {
                searchForSkoog()
            }
        case .msgPropertyChanged:
            print("Property changed!")
        case .msgThruConnectionsChanged:
            print("Thru connections changed!")
        case .msgSerialPortOwnerChanged:
            print("Serial port owner changed!")
        case .msgIOError:
            print("IO Error!")
        }
    }
    
    public func searchForSkoog(){
        let numberOfDevices = MIDIGetNumberOfDevices()
        
        for count:Int in 0 ..< numberOfDevices {
            let midiDevice = MIDIGetDevice(count)
            var unmanagedProperties: Unmanaged<CFPropertyList>?

            MIDIObjectGetProperties(midiDevice, &unmanagedProperties, false)
            if let midiProperties: CFPropertyList = unmanagedProperties?.takeUnretainedValue() {
                let midiDictionary = midiProperties as! NSDictionary
                if (midiDictionary.object(forKey: "name") as! String == "Skoog" && midiDictionary.object(forKey: "offline") as! Int == 0) {
                    print("Found a Skoog!")
                    skoogConnected = true
                    if skoogConnected == true {
                        delegate?.skoogConnectionStatus(connected: true)
                    }
                    // start accepting MIDI packets from Skoog
                    start()
                }
                else if (midiDictionary.object(forKey: "name") as! String == "Skoog" && midiDictionary.object(forKey: "offline") as! Int == 1){
                    print("Skoog disconnected!")
                    skoogConnected = false
                    if skoogConnected == false {
                        delegate?.skoogConnectionStatus(connected: false)
                    }
                }
                
            } else {
                print("Unable to load properties for \(count)")
            }
        }
    }

    func connectSourcesToInputPort() {
        var status = OSStatus(noErr)
        let sourceCount = MIDIGetNumberOfSources()
        print("source count \(sourceCount)")
        
        for srcIndex in 0 ..< sourceCount {
            let midiEndPoint = MIDIGetSource(srcIndex)
            status = MIDIPortConnectSource(inputPort,
                                           midiEndPoint,
                                           nil)
            if status == OSStatus(noErr) {
                print("yay connected endpoint to inputPort!")
            } else {
                print("oh crap!")
            }
        }
    }
    
    public func start(){
        var readBlock: MIDIReadBlock
        readBlock = MyMIDIReadBlock
        
        var status = OSStatus(noErr)
        
        if status == OSStatus(noErr) {
            status = MIDIInputPortCreateWithBlock(midiClient, "com.skoogmusic.MIDIInputPort" as CFString, &inputPort, readBlock)
            if status == OSStatus(noErr) {
                print("created input port")
            } else {
                print("error creating input port : \(status)")
            }
        }
        connectSourcesToInputPort()
    }

    func MyMIDIReadBlock(packetList: UnsafePointer<MIDIPacketList>, srcConnRefCon: UnsafeMutableRawPointer?) -> Swift.Void {
        let packets = packetList.pointee
        let packet:MIDIPacket = packets.packet
        var ap = UnsafeMutablePointer<MIDIPacket>.allocate(capacity: 1)
        ap.initialize(to:packet)
        
        for _ in 0 ..< packets.numPackets {
            let p = ap.pointee
            handle(p)
            ap = MIDIPacketNext(ap)
        }
    }
    
    func handle(_ packet:MIDIPacket) {
        if packet.data.1 == 102 &&
            packet.data.4 == 103 &&
            packet.data.7 == 104 &&
            packet.data.10 == 105 &&
            packet.data.13 == 106 &&
            packet.data.16 == 107 &&
            packet.data.19 == 108 {
            
            // small bit of pre-processing to get numbers 
            // from the MIDI range into a more sensible range for dealing with angles
            R1 = Double(packet.data.2) / 127.0
            T1 = Double(packet.data.5) * (360.0 / 127.0)
            R2 = Double(packet.data.8) / 127.0
            T2 = Double(packet.data.11) * (360.0 / 127.0)
            Z3 = Double(packet.data.14) / 127.0
            // once we've set our main variables the main 
            // work can be done on decoding the signal
            decodeAngle()
            // Then we can route the signals to any delegate listening for updates
            routeSignals()
        }
    }

// MARK: Helper methods
    public func setPolyMode(active: Bool) {
        if active {
            self.XYcone = 16.0
            self.ZconeZ = 16.5
            self.ZconeXY = 10
            self.polyMode = true
            
        }
        else {
            self.XYcone = 45.0
            self.ZconeZ = 45.0
            self.ZconeXY = 30
            self.polyMode = false
        }
        changeXYcone(self.XYcone)
    }
    
    public func getSkoogConnected()->Bool {
        return self.skoogConnected
    }
    
    func getMagnitude(_ x:Double, y:Double)->Double
    {
        return sqrt(x * x + y * y)
    }
    
    func radToDegrees(_ radians:Double)->Double
    {
        return radians * 180.0 / Double(M_PI)
    }
    
    func degToRadians(_ degrees:Double)->Double
    {
        return Double(M_PI) * degrees / 180
    }
    
    func gatedLowPass(_ newinput:Double, oldinput:Double, alpha:Double, LPthreshold:Double, LPgate:Double)->Double
    {
        if newinput < LPthreshold
        {
            if newinput < LPgate
            {
                return 0.0
            }
            else
            {
                return  alpha * newinput + (1.0 - alpha) * oldinput
            }
        }
        else
        {
            return newinput
        }
        //shd we just output a Double for better accuracy/smoothing?
    }
    
    func diffLowPass(_ newinput:Double, oldinput :Double, alpha:Double, LPdiff:Double)->Double
    {
        if fabs(newinput-oldinput) < LPdiff
        {
            return  alpha * newinput + (1.0 - alpha) * oldinput
        }
        else
        {
            return newinput
        }
        //shd we just output a Double for better accuracy/smoothing?
    }
    
    func variableLowPass(_ newinput:Double, oldinput:Double, alpha:Double, LPthreshold:Double, LPgate:Double)->Double
    {
        if (newinput < (3.0 * LPgate))
        {
            if (newinput < (0.25 * LPgate))
            {
                return 0.0
            }
            else
            {
                let varalpha = ((1.0 - 0.000001) * alpha / (2.75 * LPgate)) * newinput + (1 - 3.0*(1.0 - 0.000001)/2.75) * alpha
                return  varalpha * newinput + (1.0 - varalpha) * oldinput
            }
        }
        else
        {
            return alpha * newinput + (1.0 - alpha) * oldinput
        }
    }
    
    func variableDiffLowPass(_ newinput:Double, lastoutput:Double, alpha:Double, LPdiff:Double, signalstrength:Double, LPgate:Double)->Double {
        let change = fabs(newinput-lastoutput)
        if change < LPdiff {
            if change < 35.0 {
                var varalpha = alpha * change / 35.0       // Linear Alpha Variation
                
                if signalstrength < (3.0 * LPgate) {
                    varalpha *= signalstrength / (3.0 * LPgate)
                }
                return  varalpha * newinput + (1.0 - varalpha) * lastoutput
            }
            else {
                return  alpha * newinput + (1.0 - alpha) * lastoutput
            }
        }
        else {
            return newinput
        }
        //shd we just output a Double for better accuracy/smoothing?
    }
    
    // scale the input sensor values and make sure they don't exceed 1 or go below 0
    func zoom(_ x:Double, y:Double)->Double {
        if x * y > 0.999999 {
            return 0.999999
        }
        else if x * y < 0.0000001 {
            return 0.0000001
        }
        else {
            return x * y
        }
    }
    
    func changeXYcone(_ angle:Double) {
        XYcone = angle
        coneAngle = 90.0 - (2.0 * angle)
        redBlueAngle = 180.0 + angle
        redGreenAngle = 90.0 + angle
        blueYellowAngle = 270.0 + angle
        yellowGreenAngle = angle
    }
    
    func calculateResponse (amount: Double, input: Double)->Double {
        return input / (pow(2,-amount) * (1.0 - input) + input);
    }
// MARK: Decode MIDI packets
    func decodeAngle() {
//        T1 = diffLowPass(T1, oldinput: T1_old, alpha: 0.05, LPdiff: 45.0)
//        T2 = diffLowPass(T2, oldinput: T2_old, alpha: 0.05, LPdiff: 45.0)
        
        // Correction to angle for single sensor activation scenarios
        if R1 == 0.0 && R2 > 0.0 {
            T1 = T2
        }
        else if R1 > 0.0 && R2 == 0 {
            T2 = T1
        }
        
        R1 = variableLowPass(zoom(R1, y: R_zoom), oldinput: R1_old, alpha: 0.045, LPthreshold: 1.5, LPgate: zoom(0.023622, y: R_zoom))
        R2 = variableLowPass(zoom(R2, y: R_zoom), oldinput: R2_old, alpha: 0.045, LPthreshold: 1.5, LPgate: zoom(0.023622, y: R_zoom))
        
        // calculate average Rxy from sensors 1 and 2
        Rxy = Double(M_SQRT1_2) * getMagnitude(R1, y: R2)
        
//        T1 = variableDiffLowPass(T1, lastoutput: T1_old, alpha: 0.166, LPdiff: 45.0, signalstrength:Rxy, LPgate: zoom(0.023622, y: R_zoom))
        T1 = variableDiffLowPass(T1, lastoutput: T1_old, alpha: 0.12, LPdiff: 45.0, signalstrength:Rxy, LPgate: zoom(0.023622, y: R_zoom))
        
        Txy = T1
        
        
        Z3 = gatedLowPass(zoom(Z3, y: R_zoom), oldinput: Z3_old, alpha: 0.052, LPthreshold: 1.5, LPgate: z_threshold)
        
        Rxyz = getMagnitude(Rxy, y: Z3)
        
        param3 = 0.0
        param4 = 0.0
        
        if Rxyz > thresholdZoom || Rxy > thresholdZoom || Z3 > z_thresholdZoom {
            if Rxyz > thresholdZoom {
                Rxyz = (Rxyz - thresholdZoom)/(1.0 - thresholdZoom) //normalise output after taking threshold into account
            }
            if Rxy > thresholdZoom {
                Rxy = (Rxy - thresholdZoom)/(1.0 - thresholdZoom) //normalise output after taking threshold into account
            }
            if Z3 > z_thresholdZoom {
                Z3 = (Z3 - z_thresholdZoom)/(1.0 - z_thresholdZoom) //normalise output after taking threshold into account
            }
            
            Tz = radToDegrees(atan2(Z3, Rxy))
            
            if lastZone != .none {
                if (lastZone == .green_side || lastZone == .yellow_side ||
                    lastZone == .orange_red_green || lastZone == .orange_yellow_green ||
                    lastZone == .orange_red_blue || lastZone == .orange_blue_yellow ||
                    lastZone == .orange_red || lastZone == .orange_blue ||
                    lastZone == .orange_yellow || lastZone == .orange_green) && Rxy == 0.0 && Tz == 90.0
                {
                    blocking = true
                }
                
                if blocking {
                    Z3 = 0.0
                    Tz = Tz_old
                }
                Tz = diffLowPass(Tz, oldinput: Tz_old, alpha: 0.166, LPdiff: 91.0)
            }
            else {
                blocking = false
                Tz = diffLowPass(Tz, oldinput: Tz_old, alpha: 0.95, LPdiff: 91.0)
            }
            
            
            var m : Double = 0.85
            var c : Double = 23.6
            var correction_factor_XY : Double = 0.0
            
            if Txy < 180.0 || Txy > (360.0 - XYcone) && Tz < 90.0 {
                if Tz > 74.6 {
                    m = 0.2916
                    c = 58.2633
                }
                
                correction_factor_XY = sin(degToRadians(Txy)) + 1.0 * (((Txy < 45.0 || Txy > 315.0)) ? cos(degToRadians(2.0 * Txy)) : 0.0)
                
                Tz_corrected = Tz * (1.0 - correction_factor_XY * (1.0 - 1.0 / m)) - correction_factor_XY * c / m
                
                if (Tz_corrected <= 0.0) {
                    Tz_corrected    = 0.000000001
                    ZC              = 0.000000001
                }
                else if (Tz_corrected >= 78.5) { // Above 78.5 tanf blows up.
                    Tz_corrected = Tz
                    ZC = Z3
                }
                else {
                    ZC = Rxy * tan(degToRadians(Tz_corrected))
                }
                Rxyz = getMagnitude(Rxy, y: ZC)
            }
            else {
                ZC = Z3
                Tz_corrected = Tz
            }
            
            if (Tz_corrected >= 90 - ZconeZ && Tz_corrected <= 90 + ZconeZ) {
                zone = .orange_side
                param4 = -((Tz_corrected - 90) / ZconeZ)
            }
            else {
                if Tz_corrected <= ZconeXY {
                    if Txy >= 270.0 - XYcone && Txy <= 270.0 + XYcone {
                        zone = .blue_side
                        param3 = (Txy - 270.0) / XYcone
                    }
                    else if Txy > (180.0 + XYcone) && Txy < (270.0 - XYcone) {
                        zone = .red_blue
                        param3 = (Txy - redBlueAngle) / coneAngle
                    }
                    else if Txy >= 180.0 - XYcone && Txy <= 180.0 + XYcone {
                        zone = .red_side
                        param3 = (Txy - 180.0) / XYcone
                    }
                    else if Txy > (90.0 + XYcone) && Txy < (180.0 - XYcone) {
                        zone = .red_green
                        param3 = (Txy - redGreenAngle) / coneAngle
                    }
                    else if Txy >= 90.0 - XYcone && Txy <= 90.0 + XYcone {
                        zone = .green_side
                        param3 = (Txy-90.0)/XYcone
                    }
                    else if Txy > XYcone && Txy < (90.0 - XYcone) {
                        zone = .yellow_green
                        
                        param3 = (Txy - yellowGreenAngle) / coneAngle
                    }
                    else if (Txy >= 0.0 && Txy <= XYcone) || Txy >= 360.0 - XYcone {
                        zone = .yellow_side
                        param3 = Txy > 360.0 - XYcone ? (Txy - 360.0) / XYcone : Txy / XYcone
                    }
                    else {
                        zone = .blue_yellow
                        param3 = (Txy - blueYellowAngle) / coneAngle
                    }
                    param4 = (Tz_corrected) / ZconeXY
                }
                else {
                    if Txy >= 270.0 - XYcone && Txy <= 270.0 + XYcone {
                        zone = .orange_blue
                        param3 = (Txy-270.0) / XYcone
                    }
                    else if Txy > 180.0 + XYcone && Txy < 270.0 - XYcone {
                        zone = .orange_red_blue
                        param3 = (Txy - 180.0) / (90.0 - XYcone)
                    }
                    else if Txy >= 180.0 - XYcone && Txy <= 180.0 + XYcone {
                        zone = .orange_red
                        param3 = (Txy - 180.0) / XYcone
                    }
                    else if Txy > 90.0 + XYcone && Txy < 180.0 - XYcone {
                        zone = .orange_red_green
                        param3 = (Txy - 135.0) / (45.0 - XYcone)
                    }
                    else if Txy >= 90.0 - XYcone && Txy <= 90.0 + XYcone {
                        zone = .orange_green
                        param3 = (Txy - 90.0) / XYcone
                    }
                    else if (Txy > XYcone && Txy < 90.0 - XYcone) {
                        zone = .orange_yellow_green
                        param3 = (Txy - 45.0) / (45.0 - XYcone)
                    }
                    else if (Txy >= 0.0 && Txy <= XYcone) || Txy > 360.0 - XYcone {
                        zone = .orange_yellow
                        param3 = Txy > 360.0 - XYcone ? (Txy-360.0) / XYcone : (Txy) / XYcone
                    }
                    else {
                        zone = .orange_blue_yellow
                        param3 = (Txy-315.0) / (45.0-XYcone)
                    }
                    param4 = (Tz_corrected - ZconeXY) / (90.0-ZconeZ-ZconeXY) //90-ZconeZ-ZconeXY is the width of the remaining segement
                }
            }
        }
        else {
            
            zone = .none
            Rxyz = 0.0 //normalise output after taking threshold into account
            Rxy = 0.0 //normalise output after taking threshold into account
            Z3 = 0.0 //normalise output after taking threshold into account
            ZC = 0.0 //normalise output after taking threshold into account
            Tz = ZconeXY + 0.5 * (90.0 - ZconeZ - ZconeXY) //180 * atan2f(Z3, Rxy) / M_PI
            Txy = 180.0
        }
        
        Rxyz_old = Rxyz
        Rxy_old = Rxy
        R1_old = R1
        R2_old = R2
        T1_old = T1
        T2_old = T2
        Tz_old = Tz
        Z3_old = Z3
    }
// MARK: Route decoded signals back to the delegate
    func routeSignals()
    {
        let zone1 : Int = activeZones[zone.rawValue][0].rawValue
        let zone2 : Int = activeZones[zone.rawValue][1].rawValue
        let zone3 : Int = activeZones[zone.rawValue][2].rawValue
        
        let lastZone1 : Int = activeZones[lastZone.rawValue][0].rawValue
        let lastZone2 : Int = activeZones[lastZone.rawValue][1].rawValue
        let lastZone3 : Int = activeZones[lastZone.rawValue][2].rawValue
        
        currenttime = NSDate()
        let deltaT = currenttime.timeIntervalSince(lasttime as Date)
        lasttime = currenttime
        // send shut up message(s)
        if (zone != lastZone && lastZone != .none)
        {  // don't repeatedly tell lastZone to shut up if he hasn't changed
            if (lastZone1 != zone1 &&
                lastZone1 != zone2 &&
                lastZone1 != zone3)
            {
                if (lastZone1 != Zone.none.rawValue) {
                    sides[lastZone1].rawValue = 0.0
                    sides[lastZone1].value = 0.0
                    sides[lastZone1].peak?.reset()
                    delegate?.release(side: sides[lastZone1])
                    delegate?.continuous(side: sides[lastZone1])
                }
            }
            if (lastZone2 != zone1 &&
                lastZone2 != zone2 &&
                lastZone2 != zone3)
            {
                if (lastZone2 != Zone.none.rawValue) {
                    sides[lastZone2].rawValue = 0.0
                    sides[lastZone2].value = 0.0
                    sides[lastZone2].peak?.reset()
                    delegate?.release(side: sides[lastZone2])
                    delegate?.continuous(side: sides[lastZone2])
                }
            }
            if (lastZone3 != zone1 &&
                lastZone3 != zone2 &&
                lastZone3 != zone3)
            {
                if (lastZone3 != Zone.none.rawValue) {
                    sides[lastZone3].rawValue = 0.0
                    sides[lastZone3].value = 0.0
                    sides[lastZone3].peak?.reset()
                    delegate?.release(side: sides[lastZone3])
                    delegate?.continuous(side: sides[lastZone3])
                }
            }
        }
        
        bend_out = 1.0 - 0.5 * (param3 + 1.0)
        bend_in = 1.0 - bend_out
        
        let blend_in_xy = abs(sin(param3 * M_PI_2))
        let blend_out_xy = cos(param3 * M_PI_2)
        let blend_out_z = cos(param4 * M_PI_2)
        let blend_in_z = abs(sin(param4 * M_PI_2))
        
        let inputBlendOutXY = Rxy * blend_out_xy
        let inputBlendInXY = Rxy * blend_in_xy
        let inputBlendInZ = Rxyz * blend_in_z
        let inputBlendOutZ = Rxyz * blend_out_z
        let inputBlendOutXYZ = Rxyz * blend_out_z * blend_out_xy
        let inputBlendInXYZ = Rxyz * blend_out_z * blend_in_xy
        
        switch (zone)
        {
            case .none:
                if lastZone != .none
                {
                    for i in 0 ... 4 {
                        sides[i].rawValue = 0.0
                        sides[i].value = 0.0
                        sides[i].peak?.reset()
                        delegate?.release(side: sides[i])
                        delegate?.continuous(side: sides[i])
                    }
                }
            case .red_side:
                red.rawValue = Rxy
            case .blue_side:
                blue.rawValue = Rxy
            case .yellow_side:
                yellow.rawValue = Rxy
            case .green_side:
                green.rawValue = Rxy
            case .orange_side:
                orange.rawValue = Rxyz
        default:
            if polyMode
            {
                switch (zone)
                {
                    case .red_blue:
                        red.rawValue       = inputBlendOutXY
                        blue.rawValue      = inputBlendInXY
                    case .blue_yellow:
                        blue.rawValue      = inputBlendOutXY
                        yellow.rawValue    = inputBlendInXY
                    case .yellow_green:
                        yellow.rawValue    = inputBlendOutXY
                        green.rawValue     = inputBlendInXY
                    case .red_green:
                        green.rawValue     = inputBlendOutXY
                        red.rawValue       = inputBlendInXY
                    case .orange_red:
                        orange.rawValue    = inputBlendInZ
                        red.rawValue       = inputBlendOutZ
                    case .orange_blue:
                        orange.rawValue    = inputBlendInZ
                        blue.rawValue      = inputBlendOutZ
                    case .orange_yellow:
                        orange.rawValue    = inputBlendInZ
                        yellow.rawValue    = inputBlendOutZ
                    case .orange_green:
                        orange.rawValue    = inputBlendInZ
                        green.rawValue     = inputBlendOutZ
                    case .orange_red_blue:
                        orange.rawValue    = inputBlendInZ
                        red.rawValue       = inputBlendOutXYZ
                        blue.rawValue      = inputBlendInXYZ
                    case .orange_blue_yellow:
                        orange.rawValue    = inputBlendInZ
                        blue.rawValue      = inputBlendOutXYZ
                        yellow.rawValue    = inputBlendInXYZ
                    case .orange_yellow_green:
                        orange.rawValue    = inputBlendInZ
                        yellow.rawValue    = inputBlendOutXYZ
                        green.rawValue     = inputBlendInXYZ
                    case .orange_red_green:
                        orange.rawValue    = inputBlendInZ
                        red.rawValue       = inputBlendOutXYZ
                        green.rawValue     = inputBlendInXYZ
                    default: break
                }
                
                if zone1 != Zone.none.rawValue {
                    let peak = sides[zone1].peak?.detect(input: sides[zone1].rawValue,
                                                         dT: deltaT)
                    sides[zone1].value = calculateResponse(amount: sides[zone1].response,
                                                           input: sides[zone1].rawValue)
                    if peak != nil {
                        if (peak! > 0.0) {
                            delegate?.peak(side: sides[zone1],
                                           peak: calculateResponse(amount: sides[zone1].response,
                                                                                       input: peak!))
                        }
                    }
                    else {
                        delegate?.continuous(side: sides[zone1])
                    }
                }
                if zone2 != Zone.none.rawValue {
                    let peak = sides[zone2].peak?.detect(input: sides[zone2].rawValue,
                                                         dT: deltaT)
                    sides[zone2].value = calculateResponse(amount: sides[zone2].response,
                                                           input: sides[zone2].rawValue)
                    if peak != nil {
                        if (peak! > 0.0) {
                            delegate?.peak(side: sides[zone2],
                                           peak: calculateResponse(amount: sides[zone2].response,
                                                                   input: peak!))
                        }
                    }
                    else {
                        delegate?.continuous(side: sides[zone2])
                    }
                }
                if zone3 != Zone.none.rawValue {
                    let peak = sides[zone3].peak?.detect(input: sides[zone3].rawValue,
                                                         dT: deltaT)
                    sides[zone3].value = calculateResponse(amount: sides[zone3].response, input: sides[zone3].rawValue)
                    if peak != nil {
                        if (peak! > 0.0) {
                            delegate?.peak(side: sides[zone3],
                                           peak: calculateResponse(amount: sides[zone3].response,
                                                                   input: peak!))
                        }
                    }
                    else {
                        delegate?.continuous(side: sides[zone3])
                    }
                }
            }
        }
        
        
        
        if !polyMode {
            if zone.rawValue <= Zone.orange_side.rawValue {
                let peak = sides[zone.rawValue].peak?.detect(input: sides[zone.rawValue].rawValue,
                                                             dT: deltaT)
                sides[zone.rawValue].value = calculateResponse(amount: sides[zone.rawValue].response,
                                                               input: sides[zone.rawValue].rawValue)
                sides[zone.rawValue].angle = param3

                if peak != nil {
                    if (peak! > 0.0) {
                        delegate?.peak(side: sides[zone.rawValue],
                                       peak: calculateResponse(amount: sides[zone.rawValue].response,
                                                               input: peak!))
                    }
                }
                else {
                    sides[zone.rawValue].angle = param3
                    sides[zone.rawValue].blend_out_xy = blend_out_xy
                    sides[zone.rawValue].blend_in_xy = blend_in_xy

                    delegate?.continuous(side: sides[zone.rawValue])
                }
            }
        }
        lastZone = zone
    }
}
