//
//  SkoogScene.swift
//  SkoogSwitch
//
//  Created by David Skulina on 06/09/2016.
//  Copyright © 2016 Skoogmusic Ltd. All rights reserved.
//

import Foundation
import SpriteKit
import GameplayKit
import AudioToolbox

protocol SkoogSceneDelegate: class {
    func touchUp()
    func getReadyToPlay() -> Bool
//    var showCircle: Bool { get }
}

public class SkoogScene: SKScene {
    weak var sceneDelegate:SkoogSceneDelegate?
    
    public var ripple : SKShapeNode?
    public var circle : SKShapeNode?
    public var circle2 : SKShapeNode?
    public var line : SKShapeNode?
    public var label : SKLabelNode?
    public var peakLabel : SKLabelNode?

    
    public var xCalib = 0.0
    public var yCalib = 0.0
    public var zCalib = 0.0
    public var x = 0.0
    public var y = 0.0
    public var z = 0.0
    public var dt : TimeInterval = 0.0
    public var deltaX = CGFloat(0.0)
    public var deltaY = CGFloat(0.0)
    public var deltaZ = CGFloat(0.0)
    public var magnitude = CGFloat(0.0)
    public var R = CGFloat(0.0)
    public var zangle = CGFloat(0.0)
    public var angle = CGFloat(0.0)
    public var XZangle = CGFloat(0.0)
    public var YZangle = CGFloat(0.0)
    public var peak : Peak? = Peak.sharedInstance
    public var Mpeak : Double? = 0.0
    public let thresh = 3.0
    public var active = false
    public var durationMultiplier = 1.0
    public var lineWidth : CGFloat = 3.0
    public var glowAmount : CGFloat = 10.0
    public var drawCircle = true

    public let colors : [SKColor] = [   SKColor(colorLiteralRed: 218.0/255.0, green: 60.0/255.0, blue: 0.0, alpha: 1.0),         // red
                                        SKColor(colorLiteralRed: 55.0/255.0, green: 127.0/255.0, blue: 178.0/255.0, alpha: 1.0),   // blue
                                        SKColor(colorLiteralRed: 250.0/255.0, green: 217.0/255.0, blue: 0.0, alpha: 1.0),          // yellow
                                        SKColor(colorLiteralRed: 163.0/255.0, green: 203.0/255.0, blue: 0.0, alpha: 1.0),          // green
                                        SKColor(colorLiteralRed: 249.0/255.0, green: 154.0/255.0, blue: 0.0, alpha: 1.0)]          // orange
    
    private var calibTick = 0
    var counter = 0
    var timer : Timer?
    
    public required init?(coder aDecoder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    public override init(size: CGSize) {
        super.init(size: size)
        self.backgroundColor = .darkGray
        
        self.circle		= makeCircle(location:CGPoint(x:256,y:384), radius: 80, color: .lightGray)
        self.circle2	= makeCircle(location:CGPoint(x:256,y:384), radius: 10, color: .white)
        self.ripple		= makeCircle(location:CGPoint(x:256,y:384), radius: 80, color: .clear)
        self.changeCircleColor(side:0, size: 0.0)
        
        let label = SKLabelNode()
        label.position = CGPoint(x:256,y:124)
        label.color = .white
        self.addChild(label)
        self.label = label
        
        let peakLabel = SKLabelNode()
        peakLabel.position = CGPoint(x:256,y:584)
        peakLabel.color = .white
        self.addChild(peakLabel)
        self.peakLabel = peakLabel

    }
    
    
//    override func update(_ currentTime: TimeInterval) {
//        // Called before each frame is rendered
//        
//        self.label?.text = "Angle = " + String(format:"%0.0f", side.angle)
//    }
    
    
    func makeCircle(location: CGPoint, radius: CGFloat, color: SKColor) ->SKShapeNode{
        let Circle = SKShapeNode(circleOfRadius: radius ) // Size of Circle = Radius setting.
        Circle.position = location  //touch location passed from touchesBegan.
        Circle.name = "circle"
        Circle.strokeColor = color
        Circle.lineWidth = self.lineWidth
        Circle.glowWidth = 0.0
        Circle.fillColor = .clear
        self.addChild(Circle)
        return Circle
    }
    
    public func labelFade(side: Side, value: Double) {
        if let n = self.peakLabel?.copy() as! SKLabelNode? {
            if let circle = self.circle {
            n.position = CGPoint(x: circle.position.x, y: circle.position.y + 180 + CGFloat(side.index) * 35)
            }
            n.fontColor = colors[side.index]
            n.text = "Press " + String(format:"%0.2f", value)
            self.addChild(n)

			n.run(SKAction.sequence([SKAction.wait(forDuration: 0.1 * durationMultiplier),
									 SKAction.fadeOut(withDuration: 1.0 * durationMultiplier),
									 SKAction.removeFromParent()]))
        }
    }
    
    
    public func startRipple(color: UIColor, size: Double) {
        if sceneDelegate!.getReadyToPlay() {
            if (size > 0.024) {
                if let n = self.ripple?.copy() as! SKShapeNode? {
                    n.strokeColor = color

                    if let circle = self.circle {
                        n.position = CGPoint(x: circle.position.x, y: circle.position.y)
                        n.lineWidth = 3.0
                        n.glowWidth = 2.5
                        self.addChild(n)
//                        n.run(SKAction.scale(to: 1 + CGFloat(size) - CGFloat(8 * (0.1 - circle.alpha, 0.0 0.1))), duration: 1/400 * durationMultiplier))
                        n.run(SKAction.scale(to: 1 + CGFloat(size), duration: 1/40 * durationMultiplier))
                        n.run(SKAction.fadeOut(withDuration: 1.0 * durationMultiplier))
                        n.run(SKAction.sequence([SKAction.wait(forDuration: 0.1 * durationMultiplier),
                                                 SKAction.scale(to: 2.0 * (1 + CGFloat(size)), duration: 1.2 * durationMultiplier),
                                                 SKAction.removeFromParent()]))
                    }
                    else {
                        n.lineWidth = 1.0
                        n.glowWidth = 0.5
                        self.addChild(n)
                        
                        n.run(SKAction.scale(to: 1 + CGFloat(size), duration: 1/40 * durationMultiplier))
                        n.run(SKAction.fadeOut(withDuration: 1.0 * durationMultiplier))
                        n.run(SKAction.sequence([SKAction.wait(forDuration: 0.1 * durationMultiplier),
                                                 SKAction.scale(to: 10.0 * (1 + CGFloat(size)), duration: 1.2 * durationMultiplier),
                                                 SKAction.removeFromParent()]))
                    }
                }
            }
        }
    }
    
    public func changeCircleColor(side: Int, size: Double){
        let newSize = CGFloat(size * 127)
        if newSize > 0.0 {
            if let circle = self.circle {
                if sceneDelegate!.getReadyToPlay() {
                    circle.strokeColor = colors[side]
                    circle.fillColor = colors[side]
                }
            }
        }
            
        else {
            if let circle = self.circle {
                circle.strokeColor = .lightGray
                circle.fillColor = .white
            }
        }
        if let circle2 = self.circle2 {
            circle2.strokeColor = .clear
            circle2.fillColor = .lightGray
            if let circle = self.circle {
                circle2.position = CGPoint(x: circle.position.x - 2.5 * deltaX, y: circle.position.y - 2.5 * deltaY)
            }
        }
    }
    
    public func changeAlpha(side: Int, strength: Double){
        if sceneDelegate!.getReadyToPlay() {
            let newSize = CGFloat(strength * 127)
            if newSize > 0.0 {
                if let circle = self.circle {
                    circle.alpha = CGFloat(strength)
                    circle.lineWidth = self.lineWidth
                }
                if let circle2 = self.circle2 {
                    circle2.strokeColor = .clear
                    circle2.fillColor = .lightGray
                    circle2.lineWidth = self.lineWidth
                    if let circle = self.circle {
                        circle2.position = CGPoint(x: circle.position.x - 2.5 * deltaX, y: circle.position.y - 2.5 * deltaY)
                    }
                }
            }
            else {
                if let circle = self.circle {
                    circle.strokeColor = .lightGray
                    circle.fillColor = .white
                }
            }
        }
    }
    public func changeAlpha(strength: Double){
        if sceneDelegate!.getReadyToPlay() {
            let newSize = CGFloat(strength * 127)
            if newSize > 0.0 {
                if let circle = self.circle {
                    circle.alpha = CGFloat(strength)
                    circle.lineWidth = self.lineWidth
                }
                if let circle2 = self.circle2 {
                    circle2.strokeColor = .clear
                    circle2.fillColor = .lightGray
                    circle2.lineWidth = self.lineWidth
                    if let circle = self.circle {
                        circle2.position = CGPoint(x: circle.position.x - 2.5 * deltaX, y: circle.position.y - 2.5 * deltaY)
                    }
                }
            }
            else {
                if let circle = self.circle {
                    circle.strokeColor = .lightGray
                    circle.fillColor = .white
                }
            }
        }
    }

    
    
    
    public override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        for t in touches { self.touchDown(atPoint: t.location(in: self)) }
    }
    
    public override func touchesMoved(_ touches: Set<UITouch>, with event: UIEvent?) {
        for t in touches { self.touchMoved(toPoint: t.location(in: self)) }
    }
    
    public override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?) {
        for t in touches { self.touchUp(atPoint: t.location(in: self)) }
    }
    
    public override func touchesCancelled(_ touches: Set<UITouch>, with event: UIEvent?) {
        for t in touches { self.touchUp(atPoint: t.location(in: self)) }
    }
    
    
    func touchDown(atPoint pos : CGPoint) {
    }
    
    func touchMoved(toPoint pos : CGPoint) {
    }
    
    func touchUp(atPoint pos : CGPoint) {
        sceneDelegate?.touchUp()
    }
    
    public func setLineWidth(width: Double) {
        self.lineWidth = CGFloat(width)
    }
    
    public func setGlow(amount: Double) {
        self.glowAmount = CGFloat(amount)
    }
    
    
    public func changeSize(side: Int, size: Double) {
        if sceneDelegate!.getReadyToPlay() {
            let newSize = CGFloat(size * 127)
            if newSize > 0.0 {
                if let circle = self.circle {
                    circle.strokeColor = colors[side] 
                    circle.setScale(1.0 + newSize / 64)
                    circle.lineWidth = self.lineWidth // / (1.0 + 0.1 * newSize)
                    circle.glowWidth = self.glowAmount // / (1.0 + 0.1 * newSize)
                    circle.fillColor = colors[side]
                    
                }
                
                if let circle2 = self.circle2 {
                    circle2.strokeColor = .clear
                    circle2.fillColor = .lightGray
                    circle2.setScale(1 + (newSize / 64))
                    circle2.lineWidth = self.lineWidth / (1.0 + 0.1 * newSize)
                    circle2.glowWidth = 1.0 / (1.0 + 0.1 * newSize)
                    if let circle = self.circle {
                        circle2.position = CGPoint(x: circle.position.x - 2.5 * deltaX, y: circle.position.y - 2.5 * deltaY)
                    }
                }
                
                
                if let line = self.line {
                    line.yScale = 0.4 + (newSize / 25.5)
                    line.zRotation = -angle
                    line.alpha = newSize / 10
                }
            }
            else {
                if let circle = self.circle {
                    circle.setScale(1.0)
                }
                if let circle2 = self.circle2 {
                    circle2.setScale(1.0)
                    circle2.position = CGPoint(x: (self.circle?.position.x)!, y: (self.circle?.position.y)!)
                    
                }
                
                if let line = self.line {
                    line.yScale = 0.4 + (newSize / 10)
                    line.zRotation = 0.0
                    line.alpha = 0.0
                }
            }
        }
    }
    public func changeSize(size: Double) {
        if sceneDelegate!.getReadyToPlay() {
            let newSize = CGFloat(size * 127)
            if newSize > 0.0 {
                if let circle = self.circle {
                    circle.setScale(1.0 + newSize / 64)
                    circle.lineWidth = self.lineWidth // / (1.0 + 0.1 * newSize)
                    circle.glowWidth = self.glowAmount * CGFloat(size)
                }
                
                if let circle2 = self.circle2 {
                    circle2.strokeColor = .clear
                    circle2.fillColor = .lightGray
                    circle2.setScale(1 + (newSize / 64))
                    circle2.lineWidth = self.lineWidth / (1.0 + 0.1 * newSize)
                    circle2.glowWidth = 1.0 / (1.0 + 0.1 * newSize)
                    if let circle = self.circle {
                        circle2.position = CGPoint(x: circle.position.x - 2.5 * deltaX, y: circle.position.y - 2.5 * deltaY)
                    }
                }
                
                
                if let line = self.line {
                    line.yScale = 0.4 + (newSize / 25.5)
                    line.zRotation = -angle
                    line.alpha = newSize / 10
                }
            }
            else {
                if let circle = self.circle {
                    circle.setScale(1.0)
                }
                if let circle2 = self.circle2 {
                    circle2.setScale(1.0)
                    circle2.position = CGPoint(x: (self.circle?.position.x)!, y: (self.circle?.position.y)!)
                    
                }
                
                if let line = self.line {
                    line.yScale = 0.4 + (newSize / 10)
                    line.zRotation = 0.0
                    line.alpha = 0.0
                }
            }
        }
    }

}
